{
    $Id: s10.pas $
    Copyright (c) 2005 by Armin Diehl

    This program allows to read disk images from an M/A/I/ S10 written
    under BB/M

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

****************************************************************************}
{$mode objfpc}
{$longstrings on}
{$I-}
uses sysutils,getopts,send2000;

(*
Diskimage MAI S10 BBM
Note: Bit 7 is swapped on Disk, i.e. a Space is $A0 on Disk, the
      Description is for bit 7 swapped again -> Space = $20
256 Bytes/Sector
16 Sectors
2 Sides
80 Tracks
655360 Bytes

AbsSec Track Side FileOffset Content
0      0     0    0          Unknown (not used ?)
1      0     1    100        see TS10Sector1
                             Disklabel 6 Bytes
                             rest unknown
2     1     0    200         6 bytes $80
                             10 bytes $7f
                             disk description 228 bytes
                             rest unknown
3     1     1    300         up to 7ff it seems to be the index
4     2     0    400         seems to be the directory is some sort of index-file
5     2     1    500
6     3     0    600
7     3     1    700
8     4     0    800         seems to be start of directory, 26 bytes per entry, 9 Entries per Sector



.*_H8D>..
**F00
.*_FL8;.
**C17.
**P00X

!*_PDS.݀
**C21



**D20..
*_PPKRQˁ
**C20
**D02Ѐց
**D01၀


M*dERRSƀ..
*_BU0耀聀


*_CFS
*_C
NTR̀
*_DDD

*_DROP.

000009B0 7F 7F 7F 7F  80 8F 80 87  4D 2A 64 45  52 52 53 80  M*dERRS
000009C0 80 80 81 80  80 C6 80 88  93 81 80 01  80 06 80 80  ƀ..
000009D0 80 80 E1 2A  5F 42 55 30  80 84 80 80  80 81 86 E8  *_BU0
000009E0 80 80 E8 81  80 93 80 8E  80 80 80 80  80 80 80 80  聀
000009F0 80 80 80 80  80 80 80 80  80 80 80 80  80 80 80 80  

*)

(*
80 81 AC
 0  1 2c
      44d*)
var Cmd : (c_none, C_Dir, C_Extract, c_Send) = c_None;
    ImageFileName : String = '';
    OutDir : String = '';

const DirEntriesPerSector = 9;
type
  Ts10Sector1 = packed record
    ReservedSecs: array [0..2] of byte;  // 1   S=?DEC(A2$(1,3))     (00 00 02)
    DiskName    : array [0..5] of char;  // 4
    SomeId      : array [0..1] of byte;  // 10  IF A2$(10,2)=$8A9A$  (0a 1a)
    dirEntries  : array [0..2] of byte;  // 12  LET F2=?DEC(A2$(12,3));  // 00 01 2c = 300
    unknown2    : array [0..4] of byte;  // 15
    SomeId2     : byte;                  // 20  A2$(20,1)=$FFFFFF81$ (01)
    unknown3    : byte;
    unknown4    : array [0..234] of byte;
  end;
  Ts10dirEntry = packed record
    startSec   : array [0..2] of byte;
    filename   : array [0..5] of char;   // 4 bit 7 is always set (except for the 0 byte)
    filetype   : byte;  // 4=pgm,        // 10
    unknown3 : byte;                     // 14
    noRecs     : array [0..2] of char;   // 11
    RecSize    : array [0..1] of byte;   // 15
    endSec     : array [0..2] of byte;   // 17 End Sec
    unknown2   : array [0..6] of byte;
  end;
  Ts10DirSector = packed record
    dirEntries : array [0..DirEntriesPerSector-1] of Ts10DirEntry;
    unused     : array [0..21] of char;
  end;
  TbbmHdr =
    packed record
      case integer of
        0: (unknown : array [0..2] of byte;   // bb/M z80
            pname   : array [0..8] of char);
        1: (bbmid   : array [0..6] of char);   // <<BBM>>
       end;
  PbbmHdr = ^Tbbmhdr;

var ft,Disknam,Filenam : string;
    StartSector,EndSector : integer;
    DirStartSector,DirEndSector : integer;
    Sector1 : Ts10Sector1;
    rem : string;
    ip : string = DefIP;
    Port : Word = DefPort;
    DumpBasicLines : boolean = false;



procedure Help;
begin
  writeln ('Usage:');
  writeln (paramStr(0),' -d | -e | -s [--ip=] [--port=] [-o outputDir] imagefilename');
  writeln;
  writeln ('S10 reads diskimages from an MAI S10 written under BB/M');
  writeln ('-d or --dir shows disk directory');
  writeln ('-e or --extract extracts files to given (-o or --outdir) directory');
  writeln ('-s or --send sends to MAI 2000 (Special recv pgm) via TCP/IP (Terminal Server)');
  writeln ('');
  writeln ('Armin Diehl <diehl@nordrhein.de> 2005');
  Halt(1);
end;

procedure ParseCommandLine;
var C : CHAR;
    i : integer;
    GParNo: integer;
    lOpts : array [1..10] of TOption;
    lIdx  : longint;
begin
  FillChar (lOpts, sizeof (lOpts), 0);
  lOpts[1].name:='--help'; lOpts[1].has_arg:=0;
  lOpts[2].name:='--dir'; lOpts[2].has_arg:=1;
  lOpts[3].name:='--extract'; lOpts[3].has_arg:=0;
  lOpts[4].name:='--outdir'; lOpts[4].has_arg:=1;
  lOpts[5].name:='--send'; lOpts[5].has_arg:=0;
  lOpts[6].name:='--ip'; lOpts[6].has_arg:=1;
  lOpts[7].name:='--port'; lOpts[7].has_arg:=1;
  lOpts[8].name:='--dumplines'; lOpts[8].has_arg:=0;
  lIdx := 0;
  repeat
    C := GetLongOpts ('h?do:esD',@lOpts,lIdx);
    if c = #0 then
      case lIdx of
        1: c := 'h';
        2: c := 'd';
        3: c := 'e';
        4: c := 'o';
        5: c := 's';
        6: c := 'i';
        7: c := 'p';
        8: c := 'D';
      end;
    case C of
      '?','h','H' : Help;
      'd' : Cmd := c_dir;
      'e' : Cmd := c_Extract;
      'o' : begin
              OutDir := OptArg;
              if not DirectoryExists(OutDir) then
                mkdir(OutDir);
            end;
      's' : Cmd := c_send;
      'i' : IP := OptArg;
      'p' : Port := StrToInt (OptArg);
      'D' : DumpBasicLines := true;
    end;
  until C = EndOfOptions;
  if cmd = c_none then help;
  if optind=paramcount then
    ImageFileName := paramstr(optind)
  else begin
    writeln ('No image file specified');
    help;
  end;
end;


procedure swap7 (var c : array of char);
var i : integer;
begin
  for i := low(c) to High(c) do
    if c[i] > #$80 then
      byte(c[i]) := byte(c[i]) + $80;
end;

function dumphex (var x; bytes : integer) : string;
var c : byte;
    i : integer;
    p : pchar;
begin
  p := @x;
  result := '';
  for i := 1 to bytes do
  begin
    c := byte(p^) shr 4 and $0F;
    if c > 9 then
      c := byte('A') + c - 10
    else
      c:= byte('0') + c;
    result := result + char(c);
    c := byte(p^) and $0F;
    if c > 9 then
      c := byte('A') + c - 10
    else
      c:= byte('0') + c;
    result := result + char(c) + ' ';
    inc (p);
  end;
end;

function bytehex(b:byte):string;
begin
  result := dumphex(b,1);
end;


function intpower (x,y : integer) : integer;
var i : integer;
begin
  if y = 0 then result := 1 else
  begin
    result := x;
    if y > 1 then
    begin
      for i := 2 to y do
        result := result * x;
    end;
  end;
end;


function bbmdec (var b; len : byte) : longint;
var x : array [1..9999] of byte absolute b;
    bit : integer;
    i : integer;
begin
  bit := 0; result := 0;
  for i := len downto 1 do
  begin
    result := result + x[i] * intpower (2,Bit);  // (2**Bit);
    inc(bit,8);
  end;
end;

function bbmTxt (var t; len,AsciiOffset:byte):string;
var p:pchar;
    i:integer;
begin
  p:=@t;
  result := '';
  for i := 1 to len do
  begin
    if AsciiOffset = $80 then
    begin
      if p^ > #$80 then
        result := result + char(byte(p^)-$80);
    end else
      result := result + p^;
    inc(p);
  end;
  while (length(result) > 0) and (result[length(Result)] = ' ') do
    result := copy (result,1,length(result)-1);
end;

var f : file;
    offs,i,j : integer;
    d : TS10DirSector;

(*
00001F00 00 0D D5 AA  AA C6 B0 B0  00 00 00 00  00 0A 00 1E  ..**F00........
                                                     10d  30d
00001F10 40 B0 A2 A0  FB C2 C2 AF  CD FD A0 C2  C2 AF CD A0  @" {BB/M} BB/M

00001F20 C6 C9 CC C5  A0 D5 D4 C9  CC C9 D4 C9  C5 D3 00 14  FILE UTILITIES..
                                                         :  20d

00001F30 00 18 40 A2  A0 BC AA AA  C6 B0 B0 BE  A0 A0 A0 D6  ..@" <**F00>   V
            24d :
                -REM


00001F40 B2 AE B0 B0  A0 C7 C2 A0  C3 CC
                                           00 1E  00 14 40 A2  2.00 GB CL....@"
                                              30d-lin 20d-len
00001F50 A0 B1 B2 AF  B8 B2 A0 A0  C7 D2 C5 C7  A0 C2 D2 D9   12/82  GREG BRY
00001F60 C1 CE
               00 28  00 15 40 A2  A0 CC C1 CE  C7 D5 C1 C7  AN.(..@" LANGUAG
                40d     21d
00001F70 C5 A0 BA A0  C5 CE C7 CC  C9 D3 C8
                                              00  32 00 0C 40  E : ENGLISH.2..@
                                                 50d    12d
00001F80 A2 A0 D3 D5  D0 D0 CF D2  D4 A0 BA
                                              00  3C 00 16 40  " SUPPORT:.<..@
                                                 60d
00001F90 A2 A0 C9 CD  D0 CC C5 CD  C5 CE D4 C1  D4 C9 CF CE  
00001FA0 A0 BA A0 B0  B0 00 5A 00  03 40 B9 B0  00 64 00 01  .Z..@.d..
00001FB0 04 00 96 00  10 43 D1 00  04 AA DF C4  D2 E2 D5 21  ....C..!
00001FC0 FE 11 E5 25  E4 00 A0 00  19 14 D5 20  DF D1 00 12  .%.... ..
00001FD0 AA AA A0 AA  AA C6 B0 B0  A0 C1 C2 CF  D2 D4 C5 C4  ư
00001FE0 BA A0 00 AA  00 04 1F E5  25 1C 00 B4  00 04 20 E5  ....%.... 
00001FF0 25 80 00 BE  00 13 2B D4  F0 DD C8 60  E2 D2 00 01  %...+`..

swapped
00001F00 80 8D 55 2A  2A 46 30 30  80 80 80 80  80 8A 80 9E  U**F00
00001F10 C0 30 22 20  7B 42 42 2F  4D 7D 20 42  42 2F 4D 20  0" {BB/M} BB/M
00001F20 46 49 4C 45  20 55 54 49  4C 49 54 49  45 53 80 94  FILE UTILITIES
00001F30 80 98 C0 22  20 3C 2A 2A  46 30 30 3E  20 20 20 56  " <**F00>   V
00001F40 32 2E 30 30  20 47 42 20  43 4C 80 9E  80 94 C0 22  2.00 GB CL"
00001F50 20 31 32 2F  38 32 20 20  47 52 45 47  20 42 52 59   12/82  GREG BRY
00001F60 41 4E 80 A8  80 95 C0 22  20 4C 41 4E  47 55 41 47  AN" LANGUAG
00001F70 45 20 3A 20  45 4E 47 4C  49 53 48 80  B2 80 8C C0  E : ENGLISH
00001F80 22 20 53 55  50 50 4F 52  54 20 3A 80  BC 80 96 C0  " SUPPORT :
00001F90 22 20 49 4D  50 4C 45 4D  45 4E 54 41  54 49 4F 4E  " IMPLEMENTATION
00001FA0 20 3A 20 30  30 80 DA 80  83 C0 39 30  80 E4 80 81   : 00ڀ90䀁
00001FB0 84 80 16 80  90 C3 51 80  84 2A 5F 44  52 62 55 A1  .Q*_DRbU
00001FC0 7E 91 65 A5  64 80 20 80  99 94 55 A0  5F 51 80 92  ~ed U_Q
00001FD0 2A 2A 20 2A  2A 46 30 30  20 41 42 4F  52 54 45 44  ** **F00 ABORTED
00001FE0 3A 20 80 2A  80 84 9F 65  A5 9C 80 34  80 84 A0 65  : *e4e
00001FF0 A5 00 80 3E  80 93 AB 54  70 5D 48 E0  62 52 80 81  .>Tp]HbR
00002000 80 5E 62 57  A6 5D 48 8F  5E 80 48 80  BA C3 51 80  ^bW]H^HQ
00002010 86 2A 5F 48  45 41 44 62  51 80 9C 42  42 2F 4D 20  *_HEADbQBB/M
00002020 46 49 4C 45  20 55 54 49  4C 49 54 49  45 53 20 20  FILE UTILITIES
00002030 3C 2A 2A 46  30 30 3E 62  55 A1 7E 91  65 A5 64 7E  <**F00>bU~ed~
*)

(*
BBM86:
00000000 3C 3C 42 42  4D 3E 3E 3A  12 00 04 00  00 00 00 00  <<BBM>>:........
00000010 00 20 00 00  B8 00 00 64  00 07 14 D6  00 DF CA 41  . ....d....A
00000020 12 00 6E 00  08 14 D6 00  DF CB 42 12  30 00 78 00  ..n....B.0.x.
00000030 08 14 D6 00  DF CB 43 12  34 00 82 00  09 14 D6 00  ...C.4.....
00000040 DF CC 44 12  34 50 00 8C  00 09 14 D6  00 DF CC 45  D.4P.....E
00000050 12 34 56 00  96 00 0A 14  D6 00 DF CD  46 12 34 56  .4V.....F.4V
00000060 70 00 A0 00  0A 14 D6 00  DF CD 47 12  34 56 78 00  p.....G.4Vx.
00000070 AA 00 0A 14  D6 00 DF CD  46 12 34 56  78 00 B4 00  ....F.4Vx..
00000080 0A 14 D6 00  DF CD 45 12  34 56 78 00  BE 00 0A 14  ...E.4Vx....
00000090 D6 00 DF CD  44 12 34 56  78 00 C8 00  0A 14 D6 00  .D.4Vx.....
000000A0 DF CD 43 12  34 56 78 00  D2 00 0A 14  D6 00 DF CD  C.4Vx.....
000000B0 42 12 34 56  78 00 DC 00  0A 14 D6 00  DF CD 41 12  B.4Vx.....A.
000000C0 34 56 78 FF  FF 00 01 09  72 40 00 00  00 00 00 00  4Vx...r@......
*)

function bbmGetWord (var p : pchar) : word;
begin
  result := byte(p^);
  inc(p);
  result := (result shl 8) + byte(p^);
  inc(p);
end;


function getFirstREM (var f : file;
                      pstartOffs, pEndOffs : integer) : string;
var fPos : longint;
    size : integer;
    data : pointer;
    p,pEnd,linestart : pchar;
    hdr:pbbmHdr;
    lineno,linelen,w : word;
    line : ansistring;
    i : integer;
    AsciiOffset : byte;
    start : boolean;
    doDumpLine : boolean;
begin
  fPos := filepos (f);
  p := nil;
  result := '';
  try
    size := 1024;
    seek(f,pstartOffs);
    if pendOffs = 0 then
      pendOffs := filesize(f)-1;
    getmem (data,size);
    blockread(f,data^,size);
    p:=data; inc(p,sizeof(TbbmHdr));  // to first line (BB/M z80)
    hdr:=pointer(data);
    if hdr^.bbmid = '<<BBM>>' then
    begin
      inc(p,22-sizeof(hdr^));
      AsciiOffset := 0;
    end else
      AsciiOffset := $80;
    lineno := bbmGetWord (p);
    linelen:= bbmGetWord (p);
    pend := p; inc (pend,linelen);  // ptr to end of line
    if p^ = #$40 then
    begin
      inc(p,2);
      if char(byte(p^)-AsciiOffset) = '"' then inc(p);
      if char(byte(p^)-AsciiOffset) = ' ' then inc(p);
      while p < pend do
      begin
        result := result + char (byte(p^)-AsciiOffset);
        inc(p);
      end;
    end;

  finally
    seek (f,fPos);
    if p <> nil then freemem(data,size);
  end;
end;

procedure dumpBasic (fn : string;
                     pstartOffs, pEndOffs : integer; var outF : text; outFN:string);

var size : integer;
    f : file;
    data : pointer;
    p,pEnd,pTemp,linestart : pchar;
    hdr:pbbmHdr;
    lineno,linelen,w : word;
    line : ansistring;
    i : integer;
    AsciiOffset : byte;
    doDumpLine : boolean;




function bbmDecocdeVariable(var p:pchar) : string;
begin
  result := '?? ('+bytehex(byte(p^));
  inc(p);
  result := result + ' '+bytehex(byte(p^)) + ')';
  inc(p);
end;

procedure kw (const s : string);
begin
  write (outF,s+' '); inc (p);
end;

procedure kwnsp (const s : string);
begin
  write (outF,s); inc (p);
end;

procedure kwlsp (const s : string);
begin
  write (outF,' '+s); inc (p);
end;

procedure kwsp (const s : string);
begin
  write (outF,' '+s+' '); inc (p);
end;


procedure dump2eol;
begin
  while p < pend do
  begin
    write (outF,bytehex (byte(p^))+'('+char (byte(p^)-AsciiOffset)+') ');
    inc(p);
  end;
end;

procedure dumpLine;
var p : pchar;
begin
  p := linestart;
  write (outF,'< ');
  while p < pend do
  begin
    write (outF,bytehex (byte(p^))+' ');
    inc(p);
  end;
  writeln (outF,'>');
end;

function var2txt (num:word) : string;
var x,y : byte;
begin
  x := num div 16;
  y := num mod 16;
  result := char (x + byte ('A'));
  if y > 0 then
    result := result + char (byte(y)-1+byte('0'))
end;

function bcd2txt (var p : pchar; NumBytes,DigitsBeforeDec : integer) : string;
var i : integer;
    d : integer;
begin
  result := '';
  d := 0;
  for i := 1 to numBytes do
  begin

    result := result + char (((byte(p^) shr 4) and $0f) + byte('0'));
    inc(d);
    if d = DigitsBeforeDec then result := result + '.';
    result := result + char ((byte(p^) and $0f) + byte('0'));
    inc(d);
    if d = DigitsBeforeDec then result := result + '.';
    inc(p);
  end;
end;


begin
  filemode := 0;  // ReadOnly
  assign (f,fn); reset(f,1);
  seek(f,pstartOffs);
  if pendOffs = 0 then
    pendOffs := filesize(f)-1;
  getmem (data,pendOffs-pstartoffs+1);
  blockread(f,data^,pendoffs-pstartOffs+1);
  close(f);

  // save binary
  assign (f,outFN+'.bin'); rewrite(f,1);
  blockWrite(f,data^,pendoffs-pstartOffs+1);
  close(f);

  p:=data; inc(p,sizeof(TbbmHdr));  // to first line (BB/M z80)
  doDumpLine := DumpBasicLines;
  {if filenam = '!1' then
  begin
    writeln;
    doDumpLine :=true;
  end;}
  hdr:=pointer(data);
  if hdr^.bbmid = '<<BBM>>' then
  begin
    inc(p,22-sizeof(hdr^));
    AsciiOffset := 0;
  end else
    AsciiOffset := $80;
  {writeln (outF,'unknown: ',dumphex(hdr^.unknown,3));
  writeln (outF,'name   : ',hdr^.pname);}
  repeat
    linestart := p;
    lineno := bbmGetWord (p); if lineno > 9999 then exit;
    linelen := bbmGetWord (p);
    //doDumpLine :=false;
    if linelen > 0 then
    begin
      //writeln ('Linenno: ',lineno,' len: ',linelen);
      pend := p; inc (pend,linelen);  // ptr to end of line
      write (outF,lineno:4,' ');
{290 LET Z=Z+ <??
CC  43  36  52  50   ; 365.25
    ::
    ::------ 3 digits after comma ? (.250)
    :------- 4 Bytes ?
DA  D6  90  FE ; *j
12  E5  01  2C ; goto 300
->
290 LET Z=Z+365.25*J; goto 300

}
      while p < pEnd do
      begin
        case byte(p^) of
          $00 : kw ('');   // space ???
          $01 : kw ('LOAD');
          $02 : kw ('SAVE');
          $03 : kw ('FILE');
          $04 : kw ('BEGIN');
          $05 : kw ('CLEAR');
          $06 : kw ('DISABLE');
          $07 : kw ('ENABLE');
          $08 : kw ('RESERVE');
          $09 : kw ('END');
          $0A : kw ('ESCAPE');
          $0B : kw ('EXECUTE');  // 0B D4 00 - EXECUTE A$
          $0C : kw ('EXITTO');
          $0D : kw ('FLOATINGPOINT');
          $0E : kw ('FOR');
          $0F : kw ('GET');
          $10 : kw ('PUT');
          $11 : kw ('GOSUB');
          $12 : kw ('GOTO');
          $13 : kw ('IF');
          $14 : kw ('LET');
          $15 : kw ('LIST');
          $16 : kw ('NEXT');     // 1030 16  D6  80  FE  12  E5  17  70 - 1030 NEXT I; GOTO 6000
          $17 : kw ('ON');
          $18 : kw ('PRECISION');
          $19 : kw ('RELEASE');
          $1A : kw ('RESET');
          $1B : kw ('RETRY');
          $1C : kw ('RETURN');
          $1D : kw ('RUN');
          $1E : kw ('SETDAY');
          $1F : kw ('SETERR');
          $20 : kw ('SETESC');
          $21 : kw ('SETTIME');
          $22 : kw ('START');
          $23 : kw ('STOP');
          $24 : kw ('WAIT');
          $25 : kwsp ('');
          $26 : kwsp ('GOTO');  // in ON POS, same as goto
          $27 : begin           // can not be part of a compound statement
                  kw ('TABLE');
                  line := '';
                  while p < pend do
                  begin
                    line := line + inttohex (byte(p^),2); inc (p);
                  end;
                  write (outF,line);
                end;
          $28 : kw ('CLOSE');
          $29 : begin
                  kwnsp ('DEF FN');
                end;
          $2A : kw ('DELETE');
          $2B : kw ('DIM');
          $2C : kw ('DIRECT');
          $2D : kw ('EDIT');
          $2E : kw ('ENDTRACE');
          $2F : kw ('ERASE');
          $30 : kw ('EXTRACT');
          $31 : kw ('FIND');
          $32 : kw ('INDEXED');
          $33 : kw ('INPUT');
          $34 : kw ('IOLIST');
          $35 : kw ('LOCK');
          $36 : KW ('OPEN');
          $37 : kw ('PRINT');
          $38 : kw ('PROGRAM');
          $39 : kw ('READ');
          $3A : kw ('REMOVE');
          $3B : kw ('SERIAL');
          $3C : kw ('SETTRACE');
          $3D : kw ('SORT');
          $3E : kw ('UNLOCK');
          $3F : kw ('WRITE');

          $40 : begin
                  kw ('REM');
                  line := '';
                  while p < pend do
                  begin
                    line := line + char (byte(p^)-AsciiOffset);
                    inc(p);
                  end;
                  write (outF,line);
                end;
          $41 : kw ('DEBUG');
          $42 : kw ('MERGE');
          $43 : kw ('CALL');     // 9810 43  D1  00  03  D5  CE  C7  E2  D4  00  FE  1C
                                 //           Q           U   N   G   b   T
                                 // 9810 CALL "UNG",A$; RETURN *)
          $44 : kw ('EXIT');
          $45 : kw ('ENTER');
          $46 : kw ('?FormatDisk?');  // BBM86: RENAME
          $47 : kw ('ADD');
          $48 : kw ('ADDR');
          $4A : kw ('DROP');
          $4B : kw ('STRING');
          $4C : kw ('??4c');    // BBM86: CHDIR
          $4D : kw ('??4d');    // BBM86: MKDIR
          $4E : kw ('??4e');    // BBM86: RMDIR
          $4F : kw ('??4f');    // BBM86: SETDRIVE
          $50 : kw ('PREFIX');
          $61 : kwnsp (',DOM=');
          $62 : kwnsp (',END=');
          $63 : kwnsp (',ERR=');
          $64 : kwnsp (',IND=');
          $65 : kwnsp (',ISZ=');
          $66 : kwnsp (',KEY=');
          $67 : kwnsp (',LEN=');
          $68 : kwnsp (',RTY=');  // Retry, i.e. for GET
          $69 : kwnsp (',SIZ=');
          $6a : kwnsp (',TBL=');
          $6b : kwnsp (',TIM=');
          $6D : kwlsp ('IOL=');
          $80 : kwnsp ('ABS');
          $81 : kwnsp ('ASC');
          $82 : kwnsp ('CHR');
          $83 : kwnsp ('CRC');
          $84 : kwnsp ('EPT');
          $85 : kwnsp ('FN');
          $86 : kwnsp ('FPT');
          $87 : kwnsp ('HSH');
          $88 : kwnsp ('INT');
          $89 : kwnsp ('LEN');
          $8A : kwnsp ('?MOD');
          $8B : kwnsp ('NUM');
          $8C : kwnsp ('POS');
          $8D : kwnsp ('SGN');
          $8E : kwnsp ('STR');
          $8F : kwnsp ('AND');
          $90 : kwnsp ('ATH');
          $91 : kwnsp ('BIN');
          $92 : kwnsp ('DEC');
          $93 : kwnsp ('GAP');
          $94 : kwnsp ('HTA');
          $95 : kwnsp ('IOR');
          $96 : kwnsp ('LRC');
          $97 : kwnsp ('NOT');
          $98 : kwnsp ('XOR');
          $99 : kwnsp ('BSZ');
          $9A : kwnsp ('PUB');
          $9B : kwnsp ('TSK');
          $9C : kwnsp ('TCB');
          $9D : kwnsp ('DAY');
          $9E : kwnsp ('HSA');  // 2559 for a S/10 floppy, <>2559 was interpreted as harddisk (OSMONR)
          $9F : kwnsp ('SSZ');
          $A0 : kwnsp ('TIM');
          $A1 : kwnsp ('PGM');
          $A2 : kwnsp ('ERR');
          $A3 : kwnsp ('FID');
          $A4 : kwnsp ('IND');
          $A5 : kwnsp ('KEY');
          $A6 : kwnsp ('CTL');
          $A7 : kwnsp ('DSZ');
          $A8 : kwnsp ('PSZ');
          $A9 : kwnsp ('LST');
          $AA : kwnsp ('CPL');
          $AB : kwnsp ('SSN');
          $AC : kwnsp ('SYS');
          $AD : kwnsp ('KEYF');  // ?? OK ??
          $AE : kwnsp ('KEYL');  // ?? OK ??
          $AF : kwnsp ('KEYP');  // ?? OK ??
          $B0 : kwnsp ('KEYN');  // ?? OK ??

          (*$CC : begin  // bcd-number , i.e. cc 46 99 99 99  -> 999999
                                          //  cc 43 36 52 50  -> 356.25
                                          //     ::
                                          //     :-- Vorkommastellen ?
                                          //     --- Bytes including first byte ?
                end;
            *)

          $C8 : begin       // const 8 bit
                  inc(p);   // 9910 1F  C8  00    9910 SETERR 0
                  write (outF,byte(p^));
                  inc(p);
                end;
          $C9,              // const 16 Bit
          $e5 : begin       // const 16 Bit (Linennumber)
                  inc(p);   // 9920 1F  E5  26  DE   9920 SETERR 9950
                  w := bbmGetWord (p);
                  write (outF,w);
                end;
          (*$ca : begin       // some 16 bit value ?
                  inc(p);
                  w := bbmGetWord(p); inc(p);
                  write (outF,'?ca ',w,'? ');
                end;*)
          $ca..$d0
              : begin
                  i := byte(p^) - $ca + 1;  // // ca: bcd 1 byte, cb bcd 2 byte ..
                  inc(p);
                  w := byte(p^) and $0f;  // e.g. 41, dont know what the 4 is
                  if (byte(p^) and $f0) = $40 then
                  begin  // BCD
                    inc(p);
                    write (outF,bcd2txt (p,
                                         i, {NumBytes}
                                         w  {DigitsBeforeDec} ));
                  end else
                  begin  // INT CD  00  07  5B  CD  15 -> 123456789
                    inc(p);
                    write (outF,bbmdec (p^,i));
                    inc(p,i);
                  end;
                end;
          $d1 : begin
                  inc(p);
                  w := bbmGetWord (p);
                  //if p^ = #0 then  // D1 hi lo ... seems to be a string with len hi lo (16 Bit)
                  begin
                    //inc (p);
                    //w := byte (p^); inc (p);
                    line := '"';
                    for i := 1 to w do
                    begin
                      line := line + (char(byte(p^)-AsciiOffset));
                      inc(p);
                    end;
                    line := line + '"';
                    write (outF,line);
                  end;// else
                  {begin
                    write (outF,' ?? D1 ');
                    dump2eol;
                    doDumpLine := true;
                  end;}
                end;
          $d2 : begin
                  inc(p);
                  w := bbmGetWord (p);
                  //if p^ = #0 then  // D2 00 ll ... seems to be a string with len ll (in hex)
                  begin
                    //inc (p);
                    //w := byte (p^); inc (p);
                    line := '$';
                    for i := 1 to w do
                    begin
                      line := line + IntToHex (byte(p^)-AsciiOffset,2);
                      inc(p);
                    end;
                    line := line + '$';
                    write (outF,line);
                  end;{ else
                  begin
                    write (outF,' ?? D2 ');
                    dump2eol;
                  end; }
                end;
          $D4,
          $D5 : begin    // a string variable, A$ -> D4 00  T$ D5 30
                  w := bbmGetWord (p);
                  write (outF,var2txt(w-$D400)+'$');
                end;
          $D6,
          $D7 : begin    // a numeric variable, I -> D6 80
                  w := bbmGetWord (p);
                  write (outF,var2txt(w-$D600));
                end;
          $D8 : kwnsp ('+');
          $D9 : kwnsp ('-');
          $DA : kwnsp ('*');
          $DB : kwsp ('/');
          $DC : kwsp ('^');
          $DD : kwnsp ('(');
          $DE : kwnsp (')');
          $DF : kwnsp ('=');

          $E0 : kwsp ('TO');
          $E1 : kwsp ('STEP');
          $E2 : kwnsp (',');
          $E3 : kwsp ('THEN');
          $E4 : kwsp ('ELSE');
          $E7 : kwnsp ('=');
          $E8 : kwsp ('<>');
          $E9 : kwnsp ('<');
          $EA : kwnsp ('>');
          $EB : kwsp ('<=');
          $EC : kwsp ('>=');
          $ED : kwsp ('AND');
          $EE : kwsp ('OR');
          $EF : kwsp ('??ef');
          $F1 : kwnsp (':');

          $F3 : begin
                  inc(p); line := #39 + char(byte(p^)-AsciiOffset);
                  inc(p); line := line + char(byte(p^)-AsciiOffset) + #39;
                  inc(p);
                  write(outF,line);
                end;
          $F5 : kwnsp ('*');  // used in input (and may be IOLIST)
          $F2 : kwnsp (':');
          $F4 : kwnsp ('@');
          $F7 : kwnsp ('(ALL)');
          $F6 : kw ('RECORD');
          $FE : kw (';')

          else
            begin
              pTemp := p; line := ' <?? ';
              while p < pEnd do
              begin
                line := line + bytehex(byte(p^)) + ' ';
                inc(p);
              end;
              line := line + '>';
              writeln (outF,line);
              writeln (line);
              p := linestart; line := '';
              while p < pEnd do
              begin
                line := line + bytehex(byte(p^)) + ' ';
                inc(p);
              end;
              writeln (outF,line);
              writeln (line);

              (*
              write ('':5);
              p := pTemp; line := '';
              while p < pEnd do
              begin
                if p^ >= #$A0 then  // space = 20h
                  line := line + char(byte(p^)-AsciiOffset) else
                  line := line + ' ';
                line := line + '   ';
                inc(p);
              end;
              write (line);
              *)
            end;

        end;
      end;
      writeln (outF);
      if doDumpLine then dumpLine;
      //readln;
    end;
  until (linelen = 0) (* or (p^ = #$9) *) ;  // 9=END
end;


function MakeFileName(name,diskname : string) : string;
var i : integer;
begin
  if outDir = '' then
    result := diskname + DirectorySeparator + name
  else
    result := outDir + DirectorySeparator + name;
  {for i := 1 to length(result) do
    if (result[i] = '?') or (result [i] = '*') then
      result[i] := '_';}
end;


procedure savepgm (StartSec,EndSec : integer; diskname,name,Imagefn : string);
var t : text;
    p : integer;
begin
  if outDir = '' then
  begin
    diskname := MakeFileName('',diskname);
    try
      mkdir (diskname);
    except
    end;
    IOResult;
  end else
    diskname := outDir;

  name := makeFileName (name,diskname) + '.bbm';

  try
    assign (t,name); rewrite (T);
    dumpBasic (Imagefn,StartSec*256,(endSec+1)*256,t,name);
    close(t);
  except
    on e:exception do
    begin
      writeln (e.message);
      writeln ('Unable to write or process "',name,'"');
    end;
  end;
end;


{FileStructure
2 bytes RecLen
4 bytes Num Records
... records

Field Seprator is 0A, Filler is 00}
procedure saveIndexedFile (var f : file;
                    diskname,filename : string;
                    StartSector, EndSector : integer;
                    Convert80 : boolean;
                    RecLen : word; NumRecs : longint);
var outF : File;
    outFN : String;
    SecBuf: array[0..255] of char;
    i,j : integer;
begin
  outFN := makeFileName (filename,diskname);
  Assign (outF,outFn+'.indexed'); ReWrite (OutF,1);
  BlockWrite(outF,RecLen,sizeof(RecLen));               // 2 Bytes for RecLen
  BlockWrite(outF,NumRecs,sizeof(NumRecs));
  For i := StartSector to EndSector-1 do
  begin
    seek (F,i*256);
    blockRead(f,SecBuf,256);
    if Convert80 then
      for j := 0 to 255 do
        if SecBuf[j] <> #0 then  // #0 is filler
          inc(SecBuf[j],$80);
    blockWrite(OutF,SecBuf,256);
  end;
  close(OutF);
end;

(*
fro *E (Erase Disk or Directory)
  90 LET S8=SSZ(D),S0=HSA(D)
 100 DIM A2$(S8)
 110 LET H2=1;
     GET D,1,ERR=1000,A2$;
     LET S=BIN(A2$(1,3));
     IF A2$(10,2)=$FFFFFF8AFFFFFF9A$ AND A2$(20,1)=$FFFFFF81$ THEN GOTO 160
 120 INPUT (0,ERR=120)"DISKETTE IS NOT IN BB/M FORMAT, IS THIS CORRECT? (CR/N) ",X$=(""=130,"N"=10)
 130 LET S=2
 160 LET F2=BIN(A2$(12,3)),
     A$=A2$(4,6)
 170 PRINT 'LF',"CURRENT DISKETTE NAME: ",A$
 180 INPUT (0,ERR=180)"ENTER CR OR NEW NAME: ",X$=(""=200,,LEN=1,6)
 190 DIM A$(6,$FFFFFF80$);
     LET A$(1,LEN(X$))=X$
 200 IF A$(1,1)=$FFFFFF80$ THEN GOTO 180 ELSE PRINT 'LF',"CURRENT MAX NO. OF FILES: ",F2
 210 INPUT (0,ERR=210)"ENTER CR OR NEW NO.: ",X=(20150)
 220 IF X>0 THEN LET F2=X
 230 IF F2<1 THEN GOTO 210
 240 LET H2=INT((DirEntries_F2*2+4) / INT(SectorSizeS8 / 2))+1; REM F2:MaxNoOfFiles
 250 IF MOD(H2,2)=0 THEN LET H2=H2+1
 260 LET H2=H2+INT((DirEntries_F2-1) / INT(SectorSizeS8 / 26))+1+S
 270 PRINT 'LF',"ALL FILES WILL BE LOST ON DISKETTE ",D,".",'LF',"THE NEW DIRECTORY WILL OCCUPY SECT
ORS",S," THRU",H2-1,'LF'
 280 INPUT (0,ERR=270)"PARAMETERS CORRECT? (Y/N) ",X$=("Y"=290,"N"=285)
 285 ENABLE D; GOTO 20
 290 LET A2$(4,6)=A$,A2$(12,3)=ATH(F2,3),A2$(S8-72,57)=Z1$,A2$(241,16)=$FFFFFF80FFFFFF80FFFFFF80FFFF
FF80FFFFFF80FFFFFF807F7F7F7F7F7F7F7F7F7F$,A2$(17,3)=ATH(H2,3),A2$(1,3)=ATH(S,3),A2$(10,1)=$FFFFFF8A$
,A2$(11,1)=$FFFFFF9A$,A2$(15,2)=$FFFFFF80FFFFFF80$,A2$(20,1)=$FFFFFF81$
 300 PUT D,1,A2$
 310 DIM A2$(S8,$FFFFFF80$)
 320 FOR X=S TO H2-1
 330 PUT D,X,A2$
 340 NEXT X
 350 IF S4$="JD" THEN GOTO 630
 360 LET S=INT((DSZ-256) / S8); IF S<1 THEN LET S=1
 370 LET H2=H2+1; DIM A2$(S*S8,$FFFFFF80$)
 400 PUT D,H2,ERR=560,A2$
 405 PRINT @(50,21),H2:"####0"
 410 LET H2=H2+S; GOTO 400
 560 ON ERR(7,41,3,8) GOTO 570,610,610,410
*)

procedure CalcDirSectors (ReservedSectors,
                          DirEntries,
                          SectorSize : integer;
                          var DirStartSector,
                              DirEndSector : integer);
var HashSectors : integer;
    DirSectors  : integer;
begin
  HashSectors := integer((DirEntries*2+4) DIV (SectorSize DIV 2))+1;
  if HashSectors mod 2 = 0 then inc (HashSectors);

  DirSectors := ((DirEntries-1) div (SectorSize div 26));

  DirStartSector := ReservedSectors + HashSectors;
  DirEndSector := DirStartSector + DirSectors ;

end;


begin

  //dumpBasic ('bcd.bbm86',0,0,output);
  //dumpBasic ('BMCNBA.bbm86',0,0,output);
  //halt;

  //dumpBasic ('bbmutils.img',$1f00,$2cd0);
  //dumpBasic ('dienstprogramme.img',$2a00,$31ff);  // *MASK
  //dumpBasic ('dienstprogramme.img',$22a00,$25000);  // **EI
  //dumpBasic ('dienstprogramme.img',$3200,$450F);  // **A
  //dumpBasic ('dienstprogramme.img',$4510,$???);  // **B
  //dumpBasic ('dienstprogramme.img',$1ea00,$20c9f);  // **XM5
  //dumpBasic ('dienstprogramme.img',$82f00,$83e2f);  // **C
  //dumpBasic ('ad_fakt.img',$3be00,$3c20f);  // **C
  //dumpBasic ('ad_fakt.img',$16d00,$1750f);  // FAprin
  //dumpBasic ('dienstprogramme.img',$21800,$21BFF);  // INPUT
  //dumpBasic ('lss_fakt_oben.img',$8bf00,$8c73f);  // D9 - Druck Aufkleber
  //dumpBasic ('lss_fakt_oben.img',$8c900,$8f4cf);  // **VERT
  //dumpBasic ('lss_fakt_oben.img',$8f600,$900ef);  // ***B
  //dumpBasic ('lss_fakt_oben.img',$90e00,$915ff);  // ***B
  //dumpBasic ('lss_fakt_oben.img',$91600,$91ddf);  // dtest
  //dumpBasic ('lss_fakt_oben.img',$90e00,$915ff);  // Dr
  //dumpBasic ('lss_fakt_oben.img',$91A00,$91ddf);  //
  //dumpBasic ('lss_fakt_oben.img',$93200,$9444f);  // ARTIKEL
  //dumpBasic ('lss_schnellbuchung.img',$5500,$660f);  // LSS Systemstart
  //dumpBasic ('lss_schnellbuchung.img',$6700,$6DCF);  //systemstart bbm
  //dumpBasic ('lss_schnellbuchung.img',$6e00,$714f);
  //dumpBasic ('lss_schnellbuchung.img',$7c00,$886f);
  //dumpBasic ('ad_fakt.img',$1bc00,$1bdcf);  // pgmloa
  //dumpBasic ('ad_fakt.img',$38000,$391ef);  // FAA       Sector $380
  //dumpBasic ('ad_fakt.img',$3AD00,$3bd8f);  // OSMONR    Sector 3AD
  //dumpBasic ('ad_fakt.img',$3af00,$3bd8f);  // *_CON
  //dumpBasic ('ad_fakt.img',$2f00,$3f1f);  // *A print directory
  //dumpBasic ('ad_fakt.img',$5900,$5fff);  // *D BB/M rename file
  //dumpBasic ('ad_fakt.img',$6000,$68df);  // *E BB/M erase diskette
  //dumpBasic ('ad_fakt.img',$6a00,$727f);  //  *F display file information
  //dumpBasic ('ad_fakt.img',$7300,$777f);  //*G ghost task communication
  //dumpBasic ('ad_fakt.img',$7800,$86af);  // *H print direct file structure
  //dumpBasic ('ad_fakt.img',$ac00,$bb2f);  // *L  display dir
  //dumpBasic ('ad_fakt.img',$bc00,$bf4f);  // *M memory map
  //dumpBasic ('ad_fakt.img',$bf00,$d23f);  // *N disk error analysis
  //dumpBasic ('ad_fakt.img',$d300,$e3df);  // *O compute file size
  //dumpBasic ('ad_fakt.img',$fb00,$1085f);  // *P1 overlay to *P
  //dumpBasic ('ad_fakt.img',$10900,$1230f);  // *S Transfer Diskette Files
  //dumpBasic ('ad_fakt.img',$12a00,$132ff);  // XXB3
  //dumpBasic ('ad_fakt.img',$13300,$13cef);  // ICS file copy utility
  //dumpBasic ('ad_fakt.img',$13f00,$14d5f);  // *
  //dumpBasic ('ad_fakt.img',$15900,$1638f);  // plattenlabel
  //dumpBasic ('ad_fakt.img',$19500,$1b2ff);  // lieferantenstamm
  //dumpBasic ('ad_fakt.img',$1b300,$1bdcf);  // lief such
  (*
  FAA   1:00 2:01   PROGRAM    00 00 00 01 20 00 00 01 95 01 00 1d 00 43 00 00
 XXB3   1:00 2:01   PROGRAM    00 00 00 01 09 00 00 01 33 01 00 16 00 14 00 00
 $12a00 / $100 = 12a

  *)

  //halt;

  ParseCommandLine;

  offs := $700;
  filemode := 0;
  assign(f,ImageFileName);
  reset(f,1);
  seek(f,256);
  blockread (f,sector1,sizeof(sector1));

  with sector1 do
  begin
    if (SomeId[0] <> $0a) or (SomeId[1] <> $1a) or
       (SomeId2 <> $01) then
    begin
      writeln ('Image does not seem to be an S/10 BBM disk');
      halt(1);
    end;
    Disknam := bbmtxt (DiskName,6,$80);
    writeln ('Reseved Sectors: ',bbmdec(ReservedSecs,3));
    writeln ('    Dir Entries: ',bbmdec(DirEntries,3));
    writeln ('       diskname: ',DiskNam);
  end;

  CalcDirSectors (bbmdec(Sector1.ReservedSecs,3),
                  bbmdec(Sector1.DirEntries,3),
                  256,  // SectorSize
                  DirStartSector,DirEndSector);

  writeln ('DirStartSector: ',DirStartSector:4,' ofs: ',inttohex(DirStartSector*256,6));
  writeln ('     EndSector: ',DirEndSector:4,' ofs: ',inttohex((DirEndSector*256)+255,6));
  writeln ('       Sectors: ',DirendSector-DirStartSector+1);

  write   ('  File    Sectors Filetype RecSize noRecs '); if cmd = c_dir then writeln ('first REM line') else Writeln;
  writeln ('------  --------- -------- ------- ------ -------------------------------------');
  for i := DirStartSector to DirEndSector do
  begin
    seek(f,i*256);
    blockread(f,d,sizeof(d));
    for j := 0 to high(d.dirEntries) do
      with d.dirEntries[j] do
      begin
        FileNam := bbmTxt(FileName,6,$80);
        //swap7(filename);
        if filename[0] <> #0 then
          if (filename[0] <> #$ff) and (filename[0] <> #$e5) then
          begin
            StartSector := bbmdec (StartSec,3);
            EndSector := bbmdec (EndSec,3);
            rem := '';
            case filetype of
              0: ft := 'INDEXED  ';
              1: ft := 'SERIAL   ';
              2: if bbmDec(RecSize,2) = 0 then
                 ft := 'SORT     ' else
                 ft := 'DIRECT   '; // or sort if recsize=0
              //3: ft := 'STR';
              4: begin
                   ft := 'PROGRAM  ';
                   rem := getFirstREM (f,StartSector*256,endSector*256);
                 end;
             10: ft := 'DIRECTORY'
            else
                 ft := bytehex(filetype)+'?';
            end;
            //if (cmd = c_dir) or (cmd = c_send) then
            begin
              write (FileNam:6,'  ',
                StartSector:4,'-',endSector-1:4,' ',
                ft,' ',bbmDec(RecSize,2):6,' ',
                bbmDec(noRecs,3):6,' ');
              if cmd = c_dir then writeln (Copy(Rem,1,37));
                //dumphex (d.dirEntries[j],sizeof(d.dirEntries[j])));

            end;
            if cmd = c_extract then
            begin
              case filetype of
                0 : begin
                      saveIndexedFile (f,disknam,filenam,
                              StartSector, EndSector, true, bbmDec(RecSize,2),bbmDec(noRecs,3));
                      writeln ('Dumped out');
                    end;
                4 : begin
                      savepgm (StartSector,EndSector,DiskNam,FileNam,ImageFileName);
                      writeln ('Saved');
                    end else
                    writeln ('Skipped');
              end;
            end else
            if cmd = c_send then
            begin
              OutDir := '.\';
              case filetype of
                //0 : saveFile (f,disknam,filenam,
                //              StartSector, EndSector, true);
                0 : begin
                      saveIndexedFile (f,disknam,'TEMP.$$$',
                              StartSector, EndSector, true, bbmDec(RecSize,2),bbmDec(noRecs,3));
                      write ('Sending ');
                      SendIt ('I','TEMP.$$$.indexed',IP,Port,FileNam);
                      DeleteFile ('TEMP.$$$.indexed');
                      writeln;
                    end;
                4 : begin
                      {savepgm (StartSector,EndSector,DiskNam,'TEMP.$$$',ImageFileName);
                      write ('Sending ');
                      SendIt ('P','TEMP.$$$.bbm',IP,Port,FileNam);
                      DeleteFile ('TEMP.$$$.bbm');}
                      writeln;
                    end else
                      writeln ('Skipped');
              end;
            end;
          end
        {else
          writeln ('<FREE>');}
      end;
  end;
end.
